import 'package:flutter/material.dart';
import 'dart:math' as math;
import 'package:vector_math/vector_math_64.dart' show Vector3;
import '../models/graph_node.dart';

/// Simple, continously growing math class containing utility functions mostly for vector operations for now.
class MathUtil {
  static final double pi = math.pi;
  static final double twoPi = math.pi * 2.0;
  static final double halfPi = math.pi / 2.0;

  static Offset sizeToOffset(Size s) => Offset(s.width, s.height);

  static Offset offsetOf(Vector3 v3) => Offset(v3.x, v3.y);

  /// Node's position is the topLeft corner
  static Offset centreOfNode(GraphNode node) => Offset(
      node.position.dx + node.size.width * 0.5,
      node.position.dy + node.size.height * 0.5);
  static double angleOfVector(Offset direction) =>
      math.atan2(direction.dx, -direction.dy) - halfPi;

  /// Dot product of two vectors
  static double dot(Offset v1, Offset v2) => (v1.dx * v2.dx) + (v1.dy * v2.dy);

  static Offset normalize(Offset vector) {
    double len = vector.distance;
    return Offset(vector.dx / len, vector.dy / len);
  }
}
